#!/usr/bin/env python
# encoding: utf-8
"""
$Id: sdr_lut_check.py 1915 2014-03-10 16:34:41Z kathys $

Purpose: Provide support for verify that CSPP SDR products match
the corresponding IDPS products.
So
IDPS VIIRS SDR products to CSPP VIIRS SDR products
Compare LUTS used to produce products
Compare selected data sets

Input:
    The year and day hour you want compaired.

Output:
    ADL VIIRS SDR blob files, .lutasc metadata files, and HDF5 output granules will be created.

Details:
    If you input a series of granules, the software will scan the work directory.
    Thus, for N input RDR granules, you may get up to N-2 output SDR granules if they are all
    contiguous.
    It is ambiguous to provide several copies of the same granule in the work directory;
    this will result in an error abort.
    The unpacker gives each unpacking of a given granule its own

Preconditions:
    Requires ADL_HOME, CSPP_SDR_HOME, CSPP_PASS_DIR_ANC_CACHE_DIR, CSPP_PASS_DIR_ANC_HOME
    environment
    variables are set.
    Requires that any needed LD_LIBRARY_PATH is set.
    Requires that DSTATICDATA is set.


Copyright (c) 2011 University of Wisconsin Regents.
Licensed under GNU GPLv3.
"""

import logging
import os
import traceback
import datetime as dt
import glob
import datetime
import adl_query_dms
from adl_query_dms import create_cache_database, get_shipped_db, replace_file_with_link
import shutil
from adl_common import CSPP_RT_HOME
import sys
import fnmatch

from adl_common import configure_logging, check_and_convert_path, check_and_convert_env_var
from adl_common import check_anc_cache, check_existing_env_var
from adl_common import unpack, ingmsd
from adl_common import ADL_VARS, initialize_work_ADL_variables
from adl_atms_data import ADL_ATMS_ANC_GLOBS
from adl_cris_data import ADL_CRIS_ANC_GLOBS
from adl_viirs_data import ADL_VIIRS_ANC_GLOBS
from datetime import timedelta
from cspp_fetch.datasets.dataset import Dataset
from cspp_fetch.cspp_sdr_entry import SDR_ANC, sdr_ancillary
from cspp_fetch.datasets.sdr_polar import sdr_polar

import pprint
LOG = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=4)


# urls we care about

ALL_LUTS = "http://jpssdb.ssec.wisc.edu/cache/archive/LUTS_V_3_0/"
DELTA_LUTS = "http://jpssdb.ssec.wisc.edu/cspp_v_3_0/ancillary/LUTS_V_3_0/"
ANCILLARY_CACHE = "http://jpssdb.ssec.wisc.edu/cspp_v_3_0/ancillary/"

TEST_ALL_LUTS = "http://jpssdb.ssec.wisc.edu/test/ancillary/LUTS_V_3_0/"
TEST_DELTA_LUTS = "http://jpssdb.ssec.wisc.edu/cspp_v_3_0_test/ancillary/LUTS_V_3_0/"
TEST_ANCILLARY_CACHE = "http://jpssdb.ssec.wisc.edu/cspp_v_3_0_test/"


def get_url_and_pattern_for_fetch(wildcard, satelite, get_lut=False, shipped=False):
    """
    Return the url with the wild card attached for use
    with fetch.
    If the environment has an alternate url use it.
    """
    if get_lut is True:
        default = DELTA_LUTS
        if shipped is True:
            default = ALL_LUTS
    else:
        default = ANCILLARY_CACHE

    base_uri = user_remote_http_url(default)

    joined = os.path.join(base_uri, wildcard)

    final = joined % str(satelite)
    LOG.debug('requested data %s' % final)
    return final


def user_remote_http_url(remote):
    """
    :return:
    """
    varname = 'JPSS_REMOTE_ANC_DIR'
    if varname in os.environ:
        value = os.environ.get(varname)
        remote = remote.replace(ANCILLARY_CACHE, value)
        LOG.info('User remote server %s' % (remote))

    return remote


def unpack_anc(files_back, output_path=None):
    """
    """
    # unpack the file
    for anc_file in files_back:
        output_dir = os.path.dirname(anc_file)
        if output_path is None:
            output_path = output_dir
        LOG.debug('Convert output_dir to %s output_path %s' %
                  (output_dir, output_path))
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        # for ADL_Unpacker
        log_dir = os.path.join(output_dir, 'log')
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        fname = os.path.basename(anc_file)
        fname = "*." + fname
        outfile = os.path.join(output_path, fname)
        there = glob.glob(outfile)

        if not len(there) == 1:
            unpack(output_path, anc_file,silent=True)
            
def no_file_in_database(afile,use_db):
    """
    """
    if use_db:
        pattern=os.path.basename(afile)
        pattern = "".join(pattern.split(".")[:-1])
        #LOG.info(pattern)
        pattern = "*"+pattern+"*"
        
        LOG.debug("pattern %s "%(pattern))
        #LOG.info('look %s'%os.path.dirname(use_db))
        for path, dirs, files in os.walk(os.path.dirname(use_db)): 
            for filename in fnmatch.filter(files, pattern ):
                LOG.debug("Found %s"% filename )
                return False
                
    return True

def files_to_dms(work_dir, files_back, to):
    """
    """
    LOG.debug('Ingest filesS %s' % (str(files_back)))
    if files_back[-1] != '':
        output = os.path.join(ADL_VARS['ING_MSD_LZ_PATH'], to)
        LOG.debug('msg output %s' % output)
        unpack_anc(files_back, output)
        ss = glob.glob(os.path.join(ADL_VARS['ING_MSD_LZ_PATH'],"*", "*.CMP_FILENAME"))
        for f in ss:
            os.remove(f)

        tle = glob.glob(os.path.join(output, 'TLE-AUX*.msd'))
        if len(tle) >= 1:
            shutil.copy(tle[0], work_dir)
            ingmsd(work_dir)

    else:
        LOG.error('No ancillary found')

def get_anc(work_dir, dataset, desired_dt, atype, allow_remote=True,use_db=None,days_back=3,how_many=1):
    """
    """
    files_back = sdr_ancillary(
        SDR_ANC.NONE, desired_dt, allow_remote=allow_remote,
        datasets=dataset,days_back=days_back,how_many=how_many)
    
    if len(files_back) == 0:
        LOG.info('No files %s for %s' % (dataset.remote,desired_dt))
    else:
        #if allow_remote is True:
        for afile in files_back:
            if no_file_in_database(afile,use_db) :
                files_to_dms(work_dir, [afile], atype)
        LOG.debug('FILES ')
        LOG.debug(files_back)

    return files_back


def retreive_cmo(work_dir, desired_dt, spcrft, allow_remote=True,use_db=None):
    """
    CrIS-Correct-Matrix-AUX_npp_20170323210340Z_20160319183543Z_ee00000000000000Z_-_cspp_dev_all-_dev.h5
    CrIS-FS-Correct-Matrix-AUX_npp_20170327194511Z_20160319183751Z_ee00000000000000Z_-_cspp_dev_all-_dev.h5
    """

    b = "CrIS-Correct-Matrix-AUX_%s_.*?.h5" % spcrft

    dataset = Dataset(remote=os.path.join(user_remote_http_url(ANCILLARY_CACHE), "%Y_%m_%d_%j", b),
                      local="%Y_%m_%d_%j",
                      period=timedelta(seconds=1),
                      offset=timedelta(hours=0),
                      )

    atype = 'ANC'
    return get_anc(work_dir, dataset, desired_dt, atype, allow_remote=allow_remote,use_db=use_db,days_back=7)


def retreive_fs_cmo(work_dir, desired_dt, spcrft, allow_remote=True,use_db=None):
    """
    CrIS-Correct-Matrix-AUX_npp_20170323210340Z_20160319183543Z_ee00000000000000Z_-_cspp_dev_all-_dev.h5
    CrIS-FS-Correct-Matrix-AUX_npp_20170327194511Z_20160319183751Z_ee00000000000000Z_-_cspp_dev_all-_dev.h5
    """

    b = "CrIS-FS-Correct-Matrix-AUX_%s_.*?.h5" % spcrft.lower()

    dataset = Dataset(remote=os.path.join(user_remote_http_url(ANCILLARY_CACHE), "%Y_%m_%d_%j", b),
                      local="%Y_%m_%d_%j",
                      period=timedelta(seconds=1),
                      offset=timedelta(hours=0)
                      )

    atype = 'ANC'
    return get_anc(work_dir, dataset, desired_dt, atype, allow_remote=allow_remote,use_db=use_db,days_back=7)


def retreive_polar(work_dir, desired_dt, spcrft, allow_remote=True,use_db=None):
    """
    """
    b = "off_USNO-PolarWander-UT1-ANC_.*?.h5"

    dataset = Dataset(remote=os.path.join(user_remote_http_url(ANCILLARY_CACHE), "%Y_%m_%d_%j", b),
                      local="%Y_%m_%d_%j",
                      period=timedelta(seconds=1),
                      offset=timedelta(hours=0)
                      )

    atype = 'ANC'
    return get_anc(work_dir, dataset, desired_dt, atype, allow_remote=allow_remote,use_db=use_db,days_back=14,how_many=2)


def retreive_rsbautocal(work_dir, desired_dt, spcrft, allow_remote=True,use_db=None):
    """
    """
    b = "VIIRS-RSBAUTOCAL-HISTORY-AUX_..._%Y%m%d.*?.h5"
    b = "VIIRS-RSBAUTOCAL-HISTORY-AUX_%s.*?_%%Y%%m%%d.*?.h5"%spcrft.lower()

    LOG.info(b)

    dataset = Dataset(remote=os.path.join(user_remote_http_url(ANCILLARY_CACHE), "%Y_%m_%d_%j", b),
                      local="%Y_%m_%d_%j",
                      period=timedelta(seconds=1),
                      offset=timedelta(hours=0),
                      )

    atype = 'ANC'
    return get_anc(work_dir, dataset, desired_dt, atype, allow_remote=allow_remote,use_db=use_db,how_many=2)


def retreive_tle(work_dir, desired_dt, spcrft, allow_remote=True,use_db=None):
    """
    """
    b = "TLE-AUX_%s.*?.h5"%spcrft.lower()

    dataset = Dataset(remote=os.path.join(user_remote_http_url(ANCILLARY_CACHE), "%Y_%m_%d_%j", b),
                      local="%Y_%m_%d_%j",
                      period=timedelta(seconds=1),
                      offset=timedelta(hours=0)
                      )

    atype = 'AUX'
    return get_anc(work_dir, dataset, desired_dt, atype, allow_remote=allow_remote,use_db=use_db,days_back=10,how_many=3)


def viirs_anc_set(work_dir, granules_to_process, ancillary_type, spcrft, allow_cache_update=True):
    """
    """
#     ancillary_strftime = {kPOLAR: "%Y%j %H%M",
#                       kTLE: "%Y%j",
#                       kGRIB: "%Y%m%d",
#                       kNISE: "%Y%j",
#                       kRSB: "%Y%j %H:%M:%S",
#                       }
    time_fmt = ancillary_strftime[ancillary_type]

    desired_time = datetime.strptime(date, time_fmt)

    needed_ancillary_files = retreive_rsbautocal(
                    work_dir, desired_time, spcrft, allow_remote=allow_cache_update)

    needed_ancillary_files = retreive_polar(
                    work_dir, desired_time, spcrft, allow_remote=allow_cache_update)

    needed_ancillary_files = retreive_tle(
                    work_dir, desired_time, spcrft, allow_remote=allow_cache_update)





def update_ancillary_cache(cspp_sdr_home, work_dir, desired_dt, spcrft, allow_remote=True, shipped=False):

    LOG.debug('update_ancillary_cache')
    if shipped is True:
        use_db = get_shipped_db(CSPP_RT_HOME)
    else:
        use_db = create_cache_database(CSPP_RT_HOME, overwrite=False)

    initialize_work_ADL_variables(work_dir, use_db)


    retreive_polar(work_dir, desired_dt, spcrft, allow_remote=allow_remote,use_db=use_db)
    retreive_cmo(work_dir, desired_dt, spcrft, allow_remote=allow_remote,use_db=use_db)
    retreive_fs_cmo(work_dir, desired_dt, spcrft, allow_remote=allow_remote,use_db=use_db)
# 
    retreive_rsbautocal(work_dir, desired_dt, spcrft, allow_remote=allow_remote,use_db=use_db)
    retreive_tle(work_dir, desired_dt, spcrft, allow_remote=allow_remote,use_db=use_db)


def prepair_lut_for_dms(work_dir, file_back):
    """
    """
    LOG.debug(os.path.basename(file_back))
    dest = os.path.join(work_dir, 'msd', 'AUX')

    if not os.path.exists(dest):
        os.makedirs(dest)

    base_file = os.path.basename(file_back)
    shutil.copyfile(file_back, os.path.join(dest, base_file))


def retreive_lut(work_dir, desired_dt,use_db, allow_remote=True, dataset=None,how_many=3):
    """
    :return:
    """
    default = "Fixme_dir"

    os.environ['CSPP_DYNAMIC_ANCIL_DIR'] = check_and_convert_env_var('CSPP_DYNAMIC_ANCIL_DIR',
                                                                     check_write=True,
                                                                     default_value=default,silent=True)
    # get the file
    LOG.debug('desired dt %s' % (str(desired_dt)))
    files_back = sdr_ancillary(
        SDR_ANC.NONE, desired_dt, allow_remote=allow_remote,
        datasets=dataset,
        days_back=1,
        how_many=how_many)

    LOG.debug('files back %d' % len(files_back))

    new_files_back=[]
    for lut in files_back:
        if no_file_in_database(lut,use_db) :
            
            if ".h5" in lut:
                unpack_anc([lut], output_path=work_dir)
            else :
                prepair_lut_for_dms(work_dir, lut)
                new_files_back.append(lut)

    if len(new_files_back) < 1:
        LOG.info("No new files at %s"%dataset[0].remote)
        #LOG.error('No %s' % dataset[0].remote)
        #raise ValueError('No %s' % dataset[0].remote)

    return list(set(new_files_back))


def luts_galore(cspp_sdr_home, work_dir, desired_dt,spcrft, allow_remote=True, shipped=False):
    """
    """
    local_directory_for_luts = "incoming_luts"
    lut_dataset = [
        Dataset(remote="",
                local=local_directory_for_luts,
                period=timedelta(seconds=1),
                offset=timedelta(hours=0),
                )
    ]

    if shipped is True:
        use_db = get_shipped_db(CSPP_RT_HOME)
        how_many=15
    else:
        use_db = create_cache_database(CSPP_RT_HOME, overwrite=False)
        how_many=1000

    initialize_work_ADL_variables(work_dir, use_db)

    all_luts = list(
         set(ADL_VIIRS_ANC_GLOBS + ADL_ATMS_ANC_GLOBS + ADL_CRIS_ANC_GLOBS))
    
    all_files_back = []

    for lutname in all_luts:
        lut_dataset[0].remote = get_url_and_pattern_for_fetch(
                lutname, spcrft, get_lut=True, shipped=shipped)
    
            # lname = lutname % 'npp'
            # lut_dataset[0].remote = pattern % lname
        LOG.debug('lut name %s' % lut_dataset[0].remote)
    
        files_back = retreive_lut(
                work_dir, desired_dt,use_db, allow_remote=allow_remote, dataset=lut_dataset,how_many=how_many)
        for a in files_back:
                all_files_back.append(a)
                
        if len(files_back) == 0:
                LOG.debug('Missed %s'%lut_dataset[0].remote)
            
    all_files_back= list(set(all_files_back))

    ingmsd(work_dir)
    

 
    # go through the incoming lut directory
    # look for the lut in the dms
    # if it is not a link create one to the incoming luts
    db_files = os.path.dirname(use_db)

    for lutaux in all_files_back:
        lookfor = os.path.basename(lutaux)
        done = False
        for root, dirs, files in os.walk(db_files):
            for f in files:
                # LOG.info('file %s %s'%(f, lookfor))
                if lookfor in f:
                    LOG.debug('found %s' % f)
                    replace_file_with_link(lutaux, os.path.join(root, f))
                    done = True
            if done:
                break


def main():
    """ Run Viirs SDR proccessing
    """
    import argparse

    desc = """Build VIIRS SDR ProSdrViirsDBController work directory and run VIIRS SDR.
        viirs_sdr.sh wraps the python code, adl_viirs_sdr.py options are passed through."""
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument('-W', '--work-dir', metavar='work_dir', default='.',
                        help='directory in which all activity occurs, defaults to current dir')

    parser.add_argument('-d', '--debug',
                        action="store_true", default=False,
                        help="always retain intermediate files")

    parser.add_argument('-t', '--time', metavar='time',
                        required=False,
                        default=None, help="time format YYYYJJJ ")

    parser.add_argument('-v', '--verbosity', action="count", default=0,
                        help='increases verbosity 1 level ERROR-WARNING-INFO-DEBUG')

    parser.add_argument('-l', '--local',
                        action="store_true", default=False,
                        help="disable download of remote ancillary data to cache")

    parser.add_argument('--package',
                        action="store_true", default=False,
                        help="Store in package db")

    ancillary_type = ['tle', 'polar', 'rsb',  'cache', 'LUTS']
    parser.add_argument('-a', '--ancillary',
                        action="store",
                        dest="ancillary",
                        type=str,
                        required=True,
                        choices=ancillary_type,
                        help='''ancillary types\n\n
                              Possible values are...
                              {}.
                           '''.format(ancillary_type.__str__()[:]))

    args = parser.parse_args()

    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    level = levels[args.verbosity if args.verbosity < 4 else 3]

    docleanup = True

    work_dir = check_and_convert_path("WORK_DIR", args.work_dir)
    d = dt.datetime.now()
    timestamp = d.isoformat()
    logname = "viirs_sdr." + timestamp + ".log"
    logfile = os.path.join(work_dir, logname)
    # configure_logging(level, FILE=logfile)
    configure_logging(level)
    if args.debug is True:
        docleanup = False

    dpe_version = check_existing_env_var('DPE_VER', default_value=None)
    LOG.info("product version: DPE_VER for VIIRS SDR  %s" % dpe_version)

    check_anc_cache(check_write=not args.local)
    LOG.info('CSPP execution work directory is %r' % work_dir)

    spcrft = 'J01'

    if not args:
        parser.print_help()
        return 1

    rc = 1
    try:
        datetime_fmt = "%Y%j %H:%M:%S"

        user_time = args.time
        if user_time is None:
            d = dt.datetime.now()
            user_time = d.strftime('%Y%j')

        the_time = user_time + " 01:01:01"
        desired_dt = datetime.datetime.strptime(the_time, datetime_fmt)

        allow_remote = not args.local

        for spcrft in [ 'npp','j01' ] :
            if args.ancillary == 'tle':
                """
                """
                rc = retreive_tle(work_dir, desired_dt, allow_remote=allow_remote)
            if args.ancillary == 'polar':
                """
                """
                rc = retreive_polar(
                    work_dir, desired_dt, allow_remote=allow_remote)
            if args.ancillary == 'rsb':
                """
                """
                rc = retreive_rsbautocal(
                    work_dir, desired_dt, allow_remote=allow_remote)
    
            if args.ancillary == 'cache':
                update_ancillary_cache(
                    CSPP_RT_HOME, work_dir, desired_dt, spcrft,
                    allow_remote=allow_remote, shipped=args.package)
    
            if args.ancillary == 'LUTS':
                luts_galore(
                    CSPP_RT_HOME, work_dir, desired_dt, spcrft,
                    allow_remote=allow_remote, shipped=args.package)

        if rc == 0 and not args.debug:
            if os.path.exists(logfile):
                os.remove(logfile)

    except Exception:
        LOG.error(traceback.format_exc())
  

    if docleanup is True:
        if os.path.exists(logfile):
            os.remove(logfile)

        dir_to_go = os.path.join(work_dir, "msd")
        if os.path.exists(dir_to_go):
            shutil.rmtree(dir_to_go)
        dir_to_go = os.path.join(work_dir, "log")
        if os.path.exists(dir_to_go):
            shutil.rmtree(dir_to_go)

        tlef = glob.glob(os.path.join(work_dir,'TLE*.*'))
        for tle in tlef: 
            os.remove(tle)
            
        tlef = glob.glob(os.path.join(work_dir,'*.msd'))
        for tle in tlef: 
            os.remove(tle)

    return rc


if __name__ == '__main__':
    sys.exit(main())
